//gleiter6_fahren.sce
//********************************************************************
//******** G L E I T E R   *******************************************
//     Ergänzung der normalen Fahrt:
//     Aufgrund der Sensorwerte bei Erreichen eines Randes richtige 
//     Kraftrichtung vorgeben.
//     Außerdem wird eine grobe Geschwindigkeitsbegrenzung eingebaut.
//*******************************************************************


//*****************************************************************************
//****** Festlegen der Größe des sichtbaren Bereichs und dessen Skalierung ****
//*****************************************************************************

clear();
clf();

XMAX=500;
YMAX=500;

figur=gcf(); // Handler der Grafik holen
figur.figure_size= [500,500];
achsen = gca();
achsen.data_bounds = [0,0;YMAX,XMAX];//Achsen-Definitionsbereich

//***********************************************************************************
//****** Sichtbare Version der Umgebung: Einzeichnen der Bahn mit Geradenstücken ****
//***********************************************************************************
plot([0,500],[0,0]);
plot([0,500],[500,500]);
plot([0,0.01],[0,0]);

plot([100,400],[100,100]);
plot([100,400],[400,400]);
plot([100,100.01],[100,400]);
plot([400,400.01],[100,400]);

plot([150,350],[150,150]);
plot([150,350],[350,350]);
plot([150,150.01],[150,350]);
plot([350,350.01],[150,350]);

//*************************************************************************************
//****** Damit der Sensor etwas "sehen" kann, wird ein Raster erzeugt, das mit der ****
//****** jeweiligen Position korrespondiert, die "bodenmatrix"                     ****
//*************************************************************************************

bodenmatrix = zeros(YMAX,XMAX);

//Großer innerer Kasten mit hellen Elementen (4) füllen.
for xx=100:400
    for yy=100:400
        bodenmatrix(yy,xx)=4;
    end
end

//Innenbereich des großen Kanstens wieder mit dunklen Elementen (0) überschreiben.
//...so ist die "Bahn" als helle Rasterpunkte entstanden.
for xx=150:350
    for yy=150:350
        bodenmatrix(yy,xx)=0;
    end
end


//***************************************************************************************
//****** Der Sensor soll einen Bereich von 7 mal 7 Rasterpunkten erfassen können und   **
//****** wird einfach als 7mal7 Matrix definiert                                       **
//****** Bei jedem Simulationsschritt wird dann später immer der Inhalt der bodenmatrix**
//****** von dem Bereich, über dem gerade der Sensor steht in die sensormatrix hinein  **
//****** kopiert.                                                                      **
//***************************************************************************************

SENSORBREITE=7;
SENSORRADIUS = round(    (SENSORBREITE-1)/2   );

sensormatrix = zeros(SENSORBREITE,SENSORBREITE);
//Kopie der Sensormatrix mit Zeilen in umgekehrter Reihenfolge extra für Darstellung:
sensormatrix_darstellung = zeros(SENSORBREITE,SENSORBREITE); 


//********************************************************************************************************
//*** Der Zustand des Sensors wird visualisiert als kleines Quadrat in 8facher Vergrößerung unten links **
//********************************************************************************************************
Matplot1(sensormatrix_darstellung,[0,0,SENSORBREITE,SENSORBREITE]); //Matrix füllt ganzen Bereich aus, Default-Colormap.
grafik = gce();
grafik.rect = [0,0,SENSORBREITE*8,SENSORBREITE*8];


m=1.0;  //Gleitereigenschaft Masse
D=0.1;  //Gleitereigenschaft dämpfendes Verhalten

t=0;
dt=0.03;

Fx=0.0;   //Antriebskraft, mit der der Gleiter manipuliert werden kann.
Fy=0.0;

//***************************************************************************************************
//Ab hier nachfolgend können Codeteile verändert werden, um verschiede Regrlkonzepte zu realisieren *
//***************************************************************************************************

x=120;
y=120;

//x=120;    //Startposition des Fahrzeugs - Die implementierten Konzepte müssen für beliebige
//y=100;    //Startpositionen und Startgeschwindigkeiten funktionieren.
vx=0.0;   //Startgeschwindigkeit
vy=0.0;

for i=1:2000

    //UNVERÄNDERLICHER TEIL
    //Integrationsschritt, unveränderlicher Teil
    xalt = x;
    yalt = y;

    ax = (Fx - D*vx)/m;
    ay = (Fy - D*vy)/m;
    vx = vx + dt*ax;
    vy = vy + dt*ay;
     x =  x + dt*vx;
     y =  y + dt*vy;
    //Ende Integrationsschritt, unveränderlicher Teil

    //Sensormatrix aktualisieren, unveränderlicher Teil:
    ix=round(x);
    iy=round(y);
    for xx=1:SENSORBREITE
        for yy=1:SENSORBREITE
            xb = xx+ix-SENSORRADIUS;
            yb = yy+iy-SENSORRADIUS;
            if (xb>0 & xb<=XMAX & yb>0 & yb<=YMAX) then 
                sensormatrix(yy,xx) = bodenmatrix(yy+iy-SENSORRADIUS,xx+ix-SENSORRADIUS);
            else
                sensormatrix(yy,xx) = 0;
            end 
        end
    end
    //Für die Darstellung Zeilen des Sensors in umgekehrter Reihenfolge nehmen, das zuerst die hohen y-Indices
    //dargestellt werden, die mit großen y-Werten korrespondieren.
    sensormatrix_darstellung = sensormatrix(SENSORBREITE:-1:1,:);


    //ENDE Sensormatrix aktualisieren, unveränderlicher Teil:


    //ENDE UNVERÄNDERLICHER TEIL

    //******************************************************************
    //Konzept für die Regelung -----------------------------------------
    
    Fx=0.0;  //..damit irgendetwas passiert
    Fy=0.0;

    if ( sum(sensormatrix)==0 ) then
        //4. Eskalationsstufe: stehenbleiben, wenn der Sensor keine Wegstücke mehr anzeigt:

        Fx = -5*vx;  //entgegen Fahrtrichtung Dämpfung stark erhöhen.
        Fy = -5*vy;
    elseif( sensormatrix(SENSORRADIUS,SENSORRADIUS)==0 ) then
        //3. Eskalationsstufe: Bei Gefahr, die Bahn zu verlassen (Sensormitte nicht mehr auf Fahrbahn)
        //                     wird eine Art Federkraft zwischen Gleiter und Bahn wirksam.
        //Richtungsvektor für die Federkraft als Schwerpunkt der hellen Felder im Sensor berechnen:

        koordx = [   -3 -2 -1 0  1  2  3
                     -3 -2 -1 0  1  2  3
                     -3 -2 -1 0  1  2  3
                     -3 -2 -1 0  1  2  3
                     -3 -2 -1 0  1  2  3
                     -3 -2 -1 0  1  2  3
                     -3 -2 -1 0  1  2  3];

        //ACHTUNG: geometrisch ist ein kleines y unten, in der Matrix ist ein kleiner Indexwert aber oben:
        koordy = [   -3 -3 -3 -3 -3 -3 -3
                     -2 -2 -2 -2 -2 -2 -2
                     -1 -1 -1 -1 -1 -1 -1
                      0  0  0  0  0  0  0 
                      1  1  1  1  1  1  1
                      2  2  2  2  2  2  2
                      3  3  3  3  3  3  3];//3.1.b 4
       
        Fx = 400 * ( sum(koordx.*sensormatrix)/sum(sensormatrix) );
        Fy = 400 * ( sum(koordy.*sensormatrix)/sum(sensormatrix) );
    elseif( sum(sensormatrix)==SENSORBREITE*SENSORBREITE*4 ) then
        //2. Eskalationsstufe:
        //Wenn sich das Fahrzeug mitten auf der Fahrbahn befindet, keine brauchbaren Sensorwerte liefert und
        //nur noch eine sehr geringe
        //Geschwindigkeit hat, diese erhalten, sonst in irgendeine Richtung einen Kraftstoß geben.

        if( vx==0 & vy==0 )
            Fx = 10;
            Fy = -10;
        elseif( vx<2 & vx>-2 & vy<2 & vy>-2 )
            Fx = vx*10.0;            
            Fy = vy*10.0;            
        end

    else
        //Keine der Eskalationsstufen greift, also normale Fahrt:
        Fx = 0;
        Fy = 0;

        //Mit Ecken arbeiten:
        if ( sensormatrix(SENSORRADIUS-2,SENSORRADIUS+2)==0 & sensormatrix(SENSORRADIUS-2,SENSORRADIUS+1)==0 &  sensormatrix(SENSORRADIUS-1,SENSORRADIUS+2)==0 ) then 
            Fx = Fx -vx*20;
            if( vy < 40 ) then 
                Fy = 11;
            end
        elseif ( sensormatrix(SENSORRADIUS+2,SENSORRADIUS+2)==0 & sensormatrix(SENSORRADIUS+2,SENSORRADIUS+1)==0 & sensormatrix(SENSORRADIUS+1,SENSORRADIUS+2)==0) then 
            Fy = Fy -vy*20;
            if( vx < -40 ) then 
                Fx = -11;
            end
        elseif ( sensormatrix(SENSORRADIUS-2,SENSORRADIUS+2)==0 & sensormatrix(SENSORRADIUS-1,SENSORRADIUS+2)==0 & sensormatrix(SENSORRADIUS-2,SENSORRADIUS+1)==0) then 
            Fx = Fx -vx*20;
            if( vy < -40 ) then 
                Fy = -11;
            end
        elseif ( sensormatrix(SENSORRADIUS-2,SENSORRADIUS-2)==0 & sensormatrix(SENSORRADIUS-2,SENSORRADIUS-1)==0 & sensormatrix(SENSORRADIUS-1,SENSORRADIUS-2)==0) then 
            Fy = Fy -vy*20;
            if( vx < 40 ) then 
                Fx = 11;
            end
        elseif ( sensormatrix(SENSORRADIUS-2,SENSORRADIUS)==0 ) then 
            Fy = Fy -vy*20;
            if( vx < 40 ) then 
                Fx = 11;
            end
        elseif ( sensormatrix(SENSORRADIUS+2,SENSORRADIUS)==0 ) then 
            Fy = Fy -vy*30;
            if( vx < -40 ) then 
                Fx = -11;
            end
        elseif ( sensormatrix(SENSORRADIUS,SENSORRADIUS+2)==0 ) then 
            Fx = Fx -vx*30;
            if( vy < 40 ) then 
                Fy = 11;
            end
        elseif ( sensormatrix(SENSORRADIUS,SENSORRADIUS-2)==0 ) then 
            Fx = Fx -vx*30;
            if( vy < -40 ) then 
                Fy = -11;
            end
        end

    end
    //ENDE Konzept für die Regelung ------------------------------------
    //******************************************************************

    //Darstellung auf dem Bildschirm (Fahrverlauf wird dargestellt.)
    sleep(10);
    plot([xalt,x],[yalt,y],'re');
    grafik.data = sensormatrix_darstellung;    
    t=t+dt;
end

